@@ -1,6 +1,9 @@
 package Tk::ToolBar;
 
 use strict;
+use Tk::Frame;
+use Tk::Balloon;
+
 use base qw/Tk::Frame/;
 use Tk::widgets qw(Frame);
 
@@ -9,16 +12,16 @@ use Carp;
 Construct Tk::Widget 'ToolBar';
 
 use vars qw/$VERSION/;
-$VERSION = 0.03;
+$VERSION = 0.04;
 
 my $edgeH = 32;
 my $edgeW = 3;
 
 my %sideToSticky = qw(
-		      bottom s
+		      top    n
 		      right  e
 		      left   w
-		      top    n
+		      bottom s
 		      );
 
 my $packIn     = '';
@@ -55,7 +58,7 @@ sub Populate {
 
     $self->{EDGE} = $edge;
 
-    $self->_packEdge;
+    $self->_packEdge($edge);
 
     $self->ConfigSpecs(
 		       -movable => [qw/METHOD movable Movable 1/],
@@ -65,6 +68,9 @@ sub Populate {
     push @allWidgets => $self;
 
     $containers{$self->{CONTAINER}} = $self;
+
+    $self->{HIGHLIGHT} = $self->{MW}->Frame(-bg => 'white');
+    $self->{BALLOON}   = $self->{MW}->Balloon;
 }
 
 sub _packSelf {
@@ -102,8 +108,9 @@ sub _packSelf {
 
 sub _packEdge {
     my $self = shift;
+    my $e    = shift;
 
-    my $e    = $self->_edge;
+    #my $e    = $self->_edge;
     my $s    = $self->{SIDE};
 
     my ($pack, $pad, $nopad);
@@ -166,8 +173,7 @@ sub _enableEdge {
 					  );
 		 } else {
 		     $self->{DUMMY} = $self->{MW}->Frame(
-							 qw/-height 30
-							 -width 50
+							 qw/
 							 -borderwidth 2
 							 -relief ridge
 							 /,
@@ -175,7 +181,26 @@ sub _enableEdge {
 								    '-y' => $y,
 								    );
 		     $self->{DUMMY}->raise;
+		     $self->{CPACKINFO} = [$self->{CONTAINER}->packInfo];
+		     $self->{SPACKINFO} = $self->manager ? [$self->packInfo] : undef;
+		     $self->{CONTAINER}->packForget;
+		     $self->packForget;
+
+		     $self->{CONTAINER}->pack(-in => $self->{DUMMY});
+		     $self->{CONTAINER}->raise;
+		     ref($_) eq 'Tk::Frame' && $_->raise for $self->{CONTAINER}->packSlaves;
 		 }
+
+		 $self->{HIGHLIGHT}->packForget;
+		 return unless $self->_whereAmI;
+
+		 # highlight the close edge.
+		 my ($op, $pp) = $self->{SIDE} =~ /top|bottom/ ?
+		     ([qw/-height 5/], [qw/-fill x/]) : ([qw/-width 5/], [qw/-fill y/]);
+
+		 $self->{HIGHLIGHT}->configure(@$op);
+		 $self->{HIGHLIGHT}->pack(-side => $self->{SIDE},
+					  @$pp);
 	     });
 
     $e->bind('<ButtonRelease-1>' => sub {
@@ -186,79 +211,116 @@ sub _enableEdge {
 	$self->{DUMMY}->destroy;
 	$self->{DUMMY} = undef;
 
-	my $p       = $e->parent;
-	my ($x, $y) = ($self->pointerx - $self->{MW}->rootx,
-		       $self->pointery - $self->{MW}->rooty);
+	return unless $self->_whereAmI(1);
+	$self->{HIGHLIGHT}->packForget;
 
-	my $w  = $self->{MW}->Width;
-	my $h  = $self->{MW}->Height;
-
-	# bound check
-	$x     = 1      if $x < 0;
-	$y     = 1      if $y < 0;
-	$x     = $w - 1 if $x > $w;
-	$y     = $h - 1 if $y > $h;
-
-	my $dx = 0;
-	my $dy = 0;
-
-	my $close = $self->{CLOSE};
-	
-	if    ($x      < $close) { $dx = $x }
-	elsif ($w - $x < $close) { $dx = $x - $w }
-
-	if    ($y      < $close) { $dy = $y }
-	elsif ($h - $y < $close) { $dy = $y - $h }
-
-	$packIn       = '';
-	if ($dx || $dy) {
-	    my $newSide;
-	    if ($dx && $dy) {
-		# which is closer?
-		if (abs($dx) < abs($dy)) {
-		    $newSide = $dx > 0 ? 'left' : 'right';
-		} else {
-		    $newSide = $dy > 0 ? 'top' : 'bottom';
-		}
-	    } elsif ($dx) {
-		$newSide = $dx > 0 ? 'left' : 'right';
-	    } else {
-		$newSide = $dy > 0 ? 'top' : 'bottom';
-	    }
-
-	    # make sure we're stickable on that side.
-	    return unless $self->{STICKY} =~ /$sideToSticky{$newSide}/;
-
-	    $self->{SIDE} = $newSide;
-	} else {
-	    # check for overlaps.
-	    for my $w (@allWidgets) {
-		next if $w == $self;
-
-		my $x1 = $w->x;
-		my $y1 = $w->y;
-		my $x2 = $x1 + $w->width;
-		my $y2 = $y1 + $w->height;
-		
-		if ($x > $x1 and $y > $y1 and $x < $x2 and $y < $y2) {
-		    $packIn = $w;
-		    last;
-		}
-	    }
-
-	    return unless $packIn;
-	    $self->{SIDE} = $packIn->{SIDE};
-	}
 	# repack everything now.
 	my @allSlaves = grep {$_ ne $e} $self->{CONTAINER}->packSlaves;
 	$_   ->packForget for $self, @allSlaves, $self->{CONTAINER};
 
 	$self->_packSelf;
-	$self->_packEdge;
+	$self->_packEdge($e);
 	$self->_packWidget($_) for @allSlaves;
     });
 }
 
+sub _whereAmI {
+    my $self = shift;
+    my $flag = shift;
+
+    my $e       = $self->_edge;
+    my $p       = $e->parent;
+    my ($x, $y) = ($self->pointerx - $self->{MW}->rootx,
+		   $self->pointery - $self->{MW}->rooty);
+
+    my $x2 = $x + $self->{CONTAINER}->width;
+    my $y2 = $y + $self->{CONTAINER}->height;
+
+    my $w  = $self->{MW}->Width;
+    my $h  = $self->{MW}->Height;
+
+    # bound check
+    $x     = 1      if $x  < 0;
+    $y     = 1      if $y  < 0;
+    $x     = $w - 1 if $x  > $w;
+    $y     = $h - 1 if $y  > $h;
+
+    $x2    = 1      if $x2 < 0;
+    $y2    = 1      if $y2 < 0;
+    $x2    = $w - 1 if $x2 > $w;
+    $y2    = $h - 1 if $y2 > $h;
+
+    my $dx = 0;
+    my $dy = 0;
+
+    my $close = $self->{CLOSE};
+	
+    if    ($x       < $close) { $dx = $x }
+    elsif ($w - $x2 < $close) { $dx = $x2 - $w }
+    #elsif ($w - $x < $close) { $dx = $x - $w }
+
+    if    ($y       < $close) { $dy = $y }
+    elsif ($h - $y2 < $close) { $dy = $y2 - $h }
+    #elsif ($h - $y < $close) { $dy = $y - $h }
+
+    $packIn       = '';
+    if ($dx || $dy) {
+	my $newSide;
+	if ($dx && $dy) {
+	    # which is closer?
+	    if (abs($dx) < abs($dy)) {
+		$newSide = $dx > 0 ? 'left' : 'right';
+	    } else {
+		$newSide = $dy > 0 ? 'top' : 'bottom';
+	    }
+	} elsif ($dx) {
+	    $newSide = $dx > 0 ? 'left' : 'right';
+	} else {
+	    $newSide = $dy > 0 ? 'top' : 'bottom';
+	}
+
+	# make sure we're stickable on that side.
+	return undef unless $self->{STICKY} =~ /$sideToSticky{$newSide}/;
+
+	$self->{SIDE} = $newSide;
+	return $newSide;
+    } elsif ($flag) {
+	# check for overlaps.
+	for my $w (@allWidgets) {
+	    next if $w == $self;
+
+	    my $x1 = $w->x;
+	    my $y1 = $w->y;
+	    my $x2 = $x1 + $w->width;
+	    my $y2 = $y1 + $w->height;
+		
+	    if ($x > $x1 and $y > $y1 and $x < $x2 and $y < $y2) {
+		$packIn = $w;
+		last;
+	    }
+	}
+
+	unless ($packIn) {
+	    # Repack everything the way it was.
+	    $self->_restoreMe;
+	    return undef;
+	}
+
+	$self->{SIDE} = $packIn->{SIDE};
+    } else {
+	return undef;
+    }
+
+    return 1;
+}
+
+sub _restoreMe {
+    my $self = shift;
+
+    $self->pack(@{$self->{SPACKINFO}}) if defined $self->{SPACKINFO};
+    $self->{CONTAINER}->pack(@{$self->{CPACKINFO}});
+}
+
 sub _disableEdge {
     my ($self, $e) = @_;
 
@@ -284,6 +346,7 @@ sub ToolButton {
 	croak "toolbutton can be only 'Button', 'Menubutton', 'Checkbutton', or 'Radiobutton'";
     }
 
+    my $m = delete $args{-tip} || '';
     my $b = $self->{CONTAINER}->$type(%args,
 				      -relief  => 'flat',
 				      );
@@ -294,6 +357,8 @@ sub ToolButton {
 
     $self->_packWidget($b);
 
+    $self->{BALLOON}->attach($b, -balloonmsg => $m) if $m;
+
     return $b;
 }
 
@@ -311,36 +376,55 @@ sub ToolLabel {
 
 sub ToolEntry {
     my $self = shift;
+    my %args = @_;
 
-    if ($self->{STICKY} =~ /[ew]/) {
-	croak "can create toolentry only on 'ns', 'n' or 's' toolbar";
-    }
-
-    my $l = $self->{CONTAINER}->Entry(@_, -width => 5);
+    my $m = delete $args{-tip} || '';
+    my $l = $self->{CONTAINER}->Entry(%args, -width => 5);
 
     push @{$self->{WIDGETS}} => $l;
 
     $self->_packWidget($l);
+    $self->{BALLOON}->attach($b, -balloonmsg => $m) if $m;
 
     return $l;
 }
 
 sub ToolLabEntry {
     my $self = shift;
+    my %args = @_;
 
     require Tk::LabEntry;
-    my $l = $self->{CONTAINER}->LabEntry(@_, -width => 5);
+    my $m = delete $args{-tip} || '';
+    my $l = $self->{CONTAINER}->LabEntry(%args, -width => 5);
 
     push @{$self->{WIDGETS}} => $l;
 
     $self->_packWidget($l);
+    $self->{BALLOON}->attach($b, -balloonmsg => $m) if $m;
 
     return $l;
 }
 
+sub separator {
+    my $self = shift;
+
+    my $sep  = $self->{CONTAINER}->Frame(qw/
+					 -borderwidth 5
+					 -relief sunken
+					 /);
+
+    push @{$self->{WIDGETS}} => $sep;
+    $self->{SEPARATORS}{$sep} = 1;
+    $self->_packWidget($sep);
+
+    return 1;
+}
+
 sub _packWidget {
     my ($self, $b) = @_;
 
+    return $self->_packEdge($b) if exists $self->{SEPARATORS}{$b};
+
     my ($side, $pad, $nopad) = $self->{SIDE} =~ /^top$|^bottom$/ ? 
 	qw/left -padx -pady/ : qw/top -pady -padx/;
 
@@ -360,7 +444,7 @@ sub _packWidget {
 	my @allSlaves = grep {$_ ne $e} $b->packSlaves;
 	$_   ->packForget for @allSlaves;
 
-	$top->_packEdge;
+	$top->_packEdge($e);
 	$top->_packWidget($_) for @allSlaves;
     }
 
@@ -374,6 +458,11 @@ sub _createButtonBindings {
     $b->bind('<Leave>' => sub { $b->configure(qw/-relief flat/) });
 }
 
+sub Button   { goto &ToolButton }
+sub Label    { goto &ToolLabel }
+sub Entry    { goto &ToolEntry }
+sub LabEntry { goto &ToolLabEntry }
+
 __END__
 
 =pod
@@ -392,11 +481,16 @@ Tk::ToolBar - A toolbar widget for Perl/Tk
         my $mw = new MainWindow;
         my $tb = $mw->ToolBar(qw/-movable 1 -side top/);
 
-        $tb->ToolButton(-text  => 'Button',
-                        -command => sub { print "hi\n" });
-        $tb->Label     (-text  => 'A Label');
-        $tb->LabEntry  (-label => 'A LabEntry',
-                        -labelPack => [-side => "left", -anchor => "w"]);
+        $tb->ToolButton  (-text  => 'Button',
+                          -tip   => 'tool tip',
+                          -command => sub { print "hi\n" });
+        $tb->ToolLabel   (-text  => 'A Label');
+        $tb->Label       (-text  => 'Another Label');
+        $tb->ToolLabEntry(-label => 'A LabEntry',
+                          -labelPack => [-side => "left",
+                                         -anchor => "w"]);
+
+        MainLoop;
 
 =head1 DESCRIPTION
 
@@ -404,13 +498,16 @@ This module implements a dockable toolbar. It is in the same spirit as the
 "short-cut" toolbars found in most major applications, such as most web browsers
 and text editors (where you find the "back" or "save" and other shortcut buttons).
 
-Buuttons of any type (regular, menu, check, radio) can be created inside this widget.
+Buttons of any type (regular, menu, check, radio) can be created inside this widget.
 You can also create Label, Entry and LabEntry widgets.
 Moreover, the ToolBar itself can be made dockable, such that it can be dragged to
-any edge of your window. Multiple ToolBars are embeddable inside each other.
+any edge of your window. Dragging is done in "real-time" so that you can see the
+contents of your ToolBar as you are dragging it. Furthermore, if you are close to
+a stickable edge, a visual indicator will show up along that edge to guide you.
+Also, multiple ToolBars are embeddable inside each other.
 
 If you drag a ToolBar to within 15 pixels of an edge, it will stick to that
-edge. If, the ToolBar is further than 15 pixels away from an edge, but you
+edge. If the ToolBar is further than 15 pixels away from an edge, but you
 release it over another ToolBar widget, then it will be embedded inside the
 second ToolBar. You can "un-embed" an embedded ToolBar simply by dragging it
 out. You can change the 15 pixel limit using the B<-close> option.
@@ -454,39 +551,56 @@ the ToolBar can only stick to the north (top) or south (bottom) sides. Defaults
 =head1 WIDGET METHODS
 
 The following methods are used to create widgets that are placed inside
-the ToolBar. Widgets are ordered in the same order they are created.
+the ToolBar. Widgets are ordered in the same order they are created, left to right.
+
+For all widgets, except Labels, a tooltip can be specified via the B<-tip> option.
 
 =over 4
 
 =item I<$ToolBar>-E<gt>B<ToolButton>(?-type => I<buttonType>,? I<options>)
 
+=item I<$ToolBar>-E<gt>B<Button>(?-type => I<buttonType>,? I<options>)
+
 This method creates a new Button inside the ToolBar.
 The I<-type> option can be used to specify
 what kind of button to create. Can be on of 'Button', 'Checkbutton', 'Menubutton', or
-'Radiobutton'. Any other options will be passed directly to the constructor
+'Radiobutton'. A tooltip message can be specified via the -tip option.
+Any other options will be passed directly to the constructor
 of the button. The Button object is returned.
 
 =item I<$ToolBar>-E<gt>B<ToolLabel>(I<options>)
 
+=item I<$ToolBar>-E<gt>B<Label>(I<options>)
+
 This method creates a new Label inside the ToolBar.
 Any options will be passed directly to the constructor
 of the label. The Label object is returned.
 
 =item I<$ToolBar>-E<gt>B<ToolEntry>(I<options>)
 
+=item I<$ToolBar>-E<gt>B<Entry>(I<options>)
+
 This method creates a new Entry inside the ToolBar.
-Any options will be passed directly to the constructor
+A tooltip message can be specified via the -tip option.
+Any other options will be passed directly to the constructor
 of the entry. The Entry object is returned.
 
 =item I<$ToolBar>-E<gt>B<ToolLabEntry>(I<options>)
 
+=item I<$ToolBar>-E<gt>B<LabEntry>(I<options>)
+
 This method creates a new LabEntry inside the ToolBar.
-Any options will be passed directly to the constructor
+A tooltip message can be specified via the -tip option.
+Any other options will be passed directly to the constructor
 of the labentry. The LabEntry object is returned.
 In horizontal ToolBars, the label of the LabEntry widget
 will be packed to the left of the entry. On vertical
 ToolBars, the label will be packed on top of the entry.
 
+=item I<$ToolBar>-E<gt>B<separator>
+
+This method inserts a separator.
+
 =back
 
 =head1 BUGS
@@ -501,11 +615,28 @@ if you do.
 
 =head1 TODO
 
+I have implemented everything I wanted, and then some.
+Here are things that were requested, but are not implemented yet.
+If you want more, send me requests.
+
 =over 4
 
 =item o
 
-I've implemented everything I wanted to implement. If you want more, send me requests.
+Allow buttons to be "tied" to menu items.
+
+=item o
+
+Include some built-in pixmaps for some of the more common ToolButtons,
+like "Save", "New", "Open", etc ..
+
+=item o
+
+Allow movable and fixed separators.
+
+=item o
+
+Constrain cursor within Toplevel window.
 
 =back
 
@@ -522,9 +653,8 @@ or just stick it somewhere in @INC where perl can find it. It's in pure Perl.
 
 =head1 ACKNOWLEDGEMENTS
 
-Many thanks go to Chris Whiting who has tested this widget extensively and
-gave valuable comments and suggestions. Thanks are also due to Jack
-Dunnigan for showing me the light; Jack, you da man!
+The following people have given me helpful comments and bug reports to keep me busy:
+Chris Whiting, Jack Dunnigan and Robert Brooks.
 
 =head1 AUTHOR
 
